/* vumeter.c
 *
 * Vu-metar kontrola za Visual Basic
 * Copyright (C) Obrad Bjelic (.obj@sezam, obj@osmeh.fon.bg.ac.yu)
 *
 * Casopis "PC", br. 3. */

#include "stdio.h"
#include "string.h"
#include "windows.h"
#include "vbapi.h"
#include "vumeter.h"

/* Globalna promenljiva hInstance sadrzi handle instance ovog VBX-a. */
HINSTANCE hInstance;

/* Svakoj kontroli pridruzena je jedna struktura tipa VUMETERSTRUCT koja
 * sadrzi podatke lokalne za tu kontrolu.
 *
 * Kod vu-metra, to je promenljiva MaxLevel koja sadrzi maksimalnu
 * vrednost koju vu-metar moze da prikaze. Promenljiva Level drzi
 * trenutni nivo iscrtan na vu-metru. */
typedef struct {
    int MaxLevel, Level;
} VUMETERSTRUCT, *PVUMETERSTRUCT, FAR* LPVUMETERSTRUCT;

/* LpVuMeterDeref vraca pointer na lokalne podatke na osnovu zadatog
 * handle-a kontrole. Posto nam je to cesto potrebno, definisacemo
 * makro. */
#define LpVuMeterDeref(hctl) ((LPVUMETERSTRUCT)VBDerefControl(hctl))

/* Struktura PROPINFO definise jedan property kontrole. Bitniju podaci: Ime
 * (ovde MaxLevel), tip (ovde DT_SHORT - "int" u C-u), atribute (PF_fGetData
 * znaci da se podaci za property uzimaju direktno iz strukture
 * VUMETERSTRUCT, PF_fSetMsg znaci da VB posalje poruku VBM_SETPROPERTY nasoj
 * kontroli kada treba postaviti property (da bi smo ponovo iscrtali
 * vu-metar), PF_fSaveData znaci da se property snima i u FRM datoteku pri
 * snimanju formulara a PF_fEditable znaci da se vrednost property-a moze
 * promeniti i u design rezimu) i ofset polja u strukturi VUMETERSTRUCT. */
PROPINFO Property_MaxLevel = {
    "MaxLevel",
    DT_SHORT | PF_fGetData | PF_fSetMsg | PF_fSaveData | PF_fEditable,
    OFFSETIN(VUMETERSTRUCT, MaxLevel), 0, 0, (PSTR)NULL, 0
};

/* Imamo jos jedan property - Level. Sve je isto kao i malopre, samo
 * sto se property drugacije zove. */
PROPINFO Property_Level = {
    "Level",
    DT_SHORT | PF_fGetData | PF_fSetMsg | PF_fSaveData | PF_fEditable,
    OFFSETIN(VUMETERSTRUCT, Level), 0, 0, (PSTR)NULL, 0
};

/* Svi property-i moraju biti numerisani brojevima od 0 do n-1 (onim
 * redom kako su poredjani u VuMeter_Properties, videti dalje). */
#define IPROP_VUMETER_CTLNAME         0
#define IPROP_VUMETER_INDEX           1
#define IPROP_VUMETER_HWND            2
#define IPROP_VUMETER_LEFT            3
#define IPROP_VUMETER_TOP             4
#define IPROP_VUMETER_WIDTH           5
#define IPROP_VUMETER_HEIGHT          6
#define IPROP_VUMETER_BACKCOLOR       7
#define IPROP_VUMETER_FORECOLOR       8
#define IPROP_VUMETER_MAXLEVEL        9
#define IPROP_VUMETER_LEVEL           10

/* VuMeter_Properties je niz pokazivaca na strukture PPROPINFO kojima
 * se definisu svi property-i nase kontrole. */
PPROPINFO VuMeter_Properties[] = {
     PPROPINFO_STD_CTLNAME,
     PPROPINFO_STD_INDEX,
     PPROPINFO_STD_HWND,
     PPROPINFO_STD_LEFT,
     PPROPINFO_STD_TOP,
     PPROPINFO_STD_WIDTH,
     PPROPINFO_STD_HEIGHT,
     PPROPINFO_STD_BACKCOLOR,
     PPROPINFO_STD_FORECOLOR,
     &Property_MaxLevel,
     &Property_Level,
     (PPROPINFO)NULL
};

/* VuMeter_Events je niz pokazivaca na strukture PEVENTINFO od kojih svaka
 * definise po jedan event koji podrzava vu-metar. Posto nas vu-metar ne
 * poseduje ni jedan specifican event, ostavicemo samo PEVENTINFO_STD_CLICK -
 * "VuMeter_Click" koji nastaje kada korisnik klikne misem na vu-metar.
 * Ukoliko su potrebni neki specificni event-i, potrebno je definisati jednu
 * promenljivu tipa EVENTINFO za svaki event i u niz VuMeter_Events dodati
 * pokazivace na svaki od EVENTINFO-a. */
PEVENTINFO VuMeter_Events[] = {
    PEVENTINFO_STD_CLICK,
    (PEVENTINFO)NULL
};

/* Promenljiva VuMeterModel tipa MODEL je struktura koja sadrzi sve
 * neophodne podatke koji opisuju nasu kontrolu. */
MODEL VuMeterModel = {
    VB_VERSION, 0, (PCTLPROC)NULL, 0, 0, sizeof(VUMETERSTRUCT),
    IDBMP_VUMETER, "VuMeter", "VuMeter", (PSTR)NULL,
    VuMeter_Properties, VuMeter_Events,
    IPROP_VUMETER_CTLNAME, 0, -1
};

/* Funkcija VuMeterPaint iscrtava vu-metar na osnovu podataka kao sto
 * su boja pozadine i sijalica, maksimalnog i tekuceg nivoa i sl. */
#define BULBS 10
VOID NEAR VuMeterPaint(HCTL hctl, HWND hwnd, HDC hdc) {
    LPVUMETERSTRUCT lpVuMeterStruct;
    RECT rc, rcBulb;
    COLORREF cr;
    HBRUSH hForeground;
    int BulbSize, BulbsOn, i;

    lpVuMeterStruct = LpVuMeterDeref(hctl);
    if(lpVuMeterStruct -> MaxLevel <= 0)
        return;
    if(lpVuMeterStruct -> Level > lpVuMeterStruct -> MaxLevel)
        lpVuMeterStruct -> Level = lpVuMeterStruct -> MaxLevel;
    VBGetClientRect(hctl, &rc);
    BulbSize = (rc.right - rc.left) / BULBS;
    BulbsOn = (int)((lpVuMeterStruct -> Level * BULBS) /
        (lpVuMeterStruct -> MaxLevel));
    VBGetControlProperty(hctl, IPROP_VUMETER_FORECOLOR, &cr);
    if(cr > 0xFFFFFFL)
        cr = GetSysColor(LOWORD(cr));
    hForeground = CreateSolidBrush(cr);
    for(i = 0; i < BulbsOn; i++) {
        SetRect(&rcBulb, rc.left + i * BulbSize, rc.top,
            rc.left + ((i + 1) * BulbSize) - 1, rc.bottom);
        FillRect(hdc, &rcBulb, hForeground);
    }
    DeleteObject((HGDIOBJ)hForeground);
}

/* VuMeterCtlProc je kontrolna funkcija nase kontrole. Nju poziva Visual
 * Basic kada god zeli da nas informise da se nesto dogodilo sto nama
 * moze da bude od znacaja. Na primer, kada je potrebno iscrtati kontrolu,
 * kada korisnik klikne na nju, itd itd. */
LONG CALLBACK _export VuMeterCtlProc(HCTL hctl, HWND hwnd, USHORT msg, USHORT wp, LONG lp) {
    RECT rc;
    COLORREF crBackground;
    HBRUSH hBackground;
    PAINTSTRUCT ps;

    switch(msg) {
/* Visual Basic (odn. Windows) salje poruku WM_ERASEBKGND kada je
 * potrebno obrisati pozadinu prozora. Ocitacemo boju preko
 * property-a BackColor a zatim ofarbati celu povrsinu kontrole */
        case WM_ERASEBKGND:
            VBGetClientRect(hctl, &rc);
            VBGetControlProperty(hctl, IPROP_VUMETER_BACKCOLOR, &crBackground);
            if(crBackground > 0xFFFFFFL)
                crBackground = GetSysColor(LOWORD(crBackground));
            hBackground = CreateSolidBrush(crBackground);
            FillRect((HDC)wp, &rc, hBackground);
            DeleteObject((HGDIOBJ)hBackground);
            return 1;
/* Kontrola prima poruku WM_PAINT kada je potrebno iscrtati "korisnu"
 * povrsinu kontrole. */
        case WM_PAINT:
            if(wp != 0)
        		VuMeterPaint(hctl, hwnd, (HDC)wp);
    	    else {
        		BeginPaint(hwnd, &ps);
                VuMeterPaint(hctl, hwnd, ps.hdc);
                EndPaint(hwnd, &ps);
            }
            break;
/* Kontrola prima poruku VBM_SETPROPERTY kada je potrebno promeniti vrednost
 * nekog property-a a i ta promena treba da prouzrokuje neku akciju. U nasem
 * slucaju, promena propery-a Level i MaxLevel izaziva ponovno iscrtavanje
 * kontrole (pozivom funkcija VBInvalidateRect i VBUpdateControl). */
        case VBM_SETPROPERTY:
            switch(wp) {
                case IPROP_VUMETER_MAXLEVEL:
                case IPROP_VUMETER_LEVEL:
                    if(wp == IPROP_VUMETER_MAXLEVEL)
                        LpVuMeterDeref(hctl) -> MaxLevel = (SHORT)lp;
                    else
                        LpVuMeterDeref(hctl) -> Level = (SHORT)lp;
                    VBInvalidateRect(hctl, (LPRECT)0, TRUE);
                    VBUpdateControl(hctl);
                    return 0;
            }
            break;
    }
/* Ukoliko nam nije stigla ni jedna od poruka koja nas zanima, prepustamo
 * Visual Basic-u da je obradi na standardan nacin. */
    return VBDefControlProc(hctl, hwnd, msg, wp, lp);
}

/* Funkciju VBINITCC poziva Visual Basic kako bi ovaj VBX registrovao nove
 * kontrole (u ovom slucaju samo jednu) i ev. uradio ostale inicijalizacije
 */
BOOL CALLBACK _export VBINITCC(USHORT usVersion, BOOL fRuntime) {
    VuMeterModel.pctlproc = VuMeterCtlProc;
    return VBRegisterModel(hInstance, &VuMeterModel);
}

/* Funkciju LibMain poziva Windows u trenutku kada se DLL (odn. VBX) ucitao
 * u memoriju. U ovom trenutku mozemo inicijalizovati globalne promenljive
 * i sl. Povratna vrednost 1 govori Windows-u da je inicijalizacija uspela.
 */
int FAR PASCAL LibMain(HINSTANCE hModule, WORD wDataSeg, WORD cbHeapSize, LPSTR lpszCmdLine) {
    hInstance = hModule;
    return 1;
}
